package k.ld34.entities;

import haxe.ds.StringMap;
import k.lib.utils.Ease;
import k.lib.utils.KMath;
import k.lib.sound.SoundManager;
import k.lib.graphics.DisplayObject;
import k.lib.graphics.EntityObject;
import k.lib.graphics.Shake;

import k.ld34.Main;
import k.ld34.Stage;

import k.ld34.effects.Explosion;

#if DEBUG
import k.lib.gconsole.GameConsole;
#end

typedef GroupRules = {
	foes:Array<String>,			// Groups of entities considered as foe (damage, etc.) AND blocking (?)
	blocking:Array<String>,		// Groups of entities that cannot be crossed
	trigger:Array<String>		// Groups of entities triggering an action upon collision (bonus, etc.)
};

class Entity extends EntityObject {
	public static var GRID_SIZE:Int = 16;
	public var health:Int = -1;
	public var attack:Int = 1;

	private static var pool:StringMap<Array<Entity>>;
	private static var group_rules:StringMap<GroupRules>;
	private static var animations:StringMap<Array<String>>;
	public static var walls:StringMap<Array<Array<Bool>>>;

	// Used when rendering, to know if something changed
	// Do not forget to set it back to false after each render
	public static var needsUpdate:Bool = false;

	public static function resetAll():Void {
		pool = null;
		group_rules = null;
		animations = null;
		walls = null;
	}

	public static function defineGroup(name:String, ?foes:Array<String> = null, ?blocking:Array<String> = null, ?trigger:Array<String> = null):Void {
		if (pool == null) pool = new StringMap<Array<Entity>>();
		if (group_rules == null) group_rules = new StringMap<GroupRules>();

		group_rules.set(name, {
			foes: foes == null ? [] : foes,
			blocking: blocking == null ? [] : blocking,
			trigger: trigger == null ? [] : trigger
		});

		pool.set(name, new Array<Entity>());
	}

	public static inline function setWall(group:String, x:Int, y:Int):Void {
		if (walls == null) walls = new StringMap<Array<Array<Bool>>>();
		if (!walls.exists(group)) walls.set(group, new Array<Array<Bool>>());

		var _walls:Array<Array<Bool>> = walls.get(group);
		if (_walls[x] == null) _walls[x] = new Array<Bool>();
		_walls[x][y] = true;
	}

	public static inline function resetWalls():Void {
		walls = new StringMap<Array<Array<Bool>>>();
	}

	public static inline function resetEntities(?except:Array<Entity> = null):Void {
		if (pool == null) return;

		for (group in pool.keys()) {
			var entities:Array<Entity> = pool.get(group);
			for (e in entities) {
				if (except != null) for (ex in except) if (ex == e) continue;
				e.destroy();
			}
		}
	}

	public static function createAnimation(name:String, tiles:Array<String>):Void {
		if (animations == null) animations = new StringMap<Array<String>>();

		animations.set(name, tiles.copy());
	}

	// Group data
	// public var group(default, null):String;
	private var _parent:Stage;
	private var _foes:Array<String>;
	private var _blocking:Array<String>;
	private var _trigger:Array<String>;
	private var has_preupdate:Bool = false;

	// Animations
	private var current_animation:Array<String>;
	public var current_animation_name:String;
	private var current_step:Int;
	private var _entityID:Int;

	// Destination
	private var hasDestination:Bool;
	private var isDeadEntity:Bool;
	private var destX:Float;
	private var destY:Float;
	private var destCallback:Null<Entity->Void>;
	private var maxSpeed:Float;
	private var symetricFrom:Entity;

	public var data:Dynamic;

	public function new(parent:Stage, group:String) {
		super(-1, DisplayObject.ALIGN_CENTER);
		this.group = group;
		_entityID = parent.addEntity(this) - 1;
		_parent = parent;
		has_update = true;

		hasDestination = false;
		isDeadEntity = false;
		destX = 0.0;
		destY = 0.0;
		destCallback = null;
		maxSpeed = 1.0;
		symetricFrom = null;

		// Init data
		cx = 0;		cy = 0;		lastcy = 0;
		xr = 0.5;	yr = 0.5;
		dx = 0.0; 	dy = 0.0;
		radius = 0; // Define in subclass

		setGroup(group);

		// Initiate animation data
		current_step = 0;
		current_animation = null;
	}

	override public function _needsUpdate():Void {
		Entity.needsUpdate = true;
	}

	public function dealDamage(amount:Int, angle:Int):Void {
		if (isDeadEntity) return;
		if (y < 0) return;

		if (health >= 0) {
			health -= amount;

			// Anim hit
			shakingColorFilter = Shake.FILTER_WHITE;
			shake(Shake.SHAKE_NEUTRAL);

			// Explosion
			var explosion:Explosion = new Explosion(_parent);
			explosion.setCoordinates(
				x + Main.random.intBetween(Std.int(-1 * radius), Std.int(radius)),
				y + Main.random.intBetween(Std.int(-1 * radius), Std.int(radius))
			);

			if (health <= 0) {
				var explosion:Explosion = new Explosion(_parent, 2);
				explosion.setCoordinates(
					x + Main.random.intBetween(Std.int(-1 * radius), Std.int(radius)),
					y + Main.random.intBetween(Std.int(-1 * radius), Std.int(radius))
				);

				var explosion:Explosion = new Explosion(_parent, 3);
				explosion.setCoordinates(
					x + Main.random.intBetween(Std.int(-1 * radius), Std.int(radius)),
					y + Main.random.intBetween(Std.int(-1 * radius), Std.int(radius))
				);

				var explosion:Explosion = new Explosion(_parent, 5);
				explosion.setCoordinates(
					x + Main.random.intBetween(Std.int(-1 * radius), Std.int(radius)),
					y + Main.random.intBetween(Std.int(-1 * radius), Std.int(radius))
				);

				if (group == "player") Main._stage.shakingColorFilter = Shake.FILTER_SOFT;
				else Main._stage.shakingColorFilter = null;

				if (group == "player" || group == "foes") Main._stage.shake(Shake.SHAKE_NEUTRAL);

				destroy();
			}
		}
	}

	public function setGroup(group:String):Void {
		// Add to the pool of entities
		if (pool != null && pool.exists(group)) {
			this.group = group;
			pool.get(group).push(this);

			// Get rules for this group of entities
			var rules:GroupRules = group_rules.get(group);
			_foes = rules.foes;
			_blocking = rules.blocking;
			_trigger = rules.trigger;
		} else {
			#if DEBUG
			GameConsole.log("Group " + group + " is not defined.");
			#end
		}
	}

	public function setCoordinates(x:Float, y:Float):Void {
		this.x = Std.int(x);
		this.y = Std.int(y);
		cx = Std.int(x / GRID_SIZE);
		cy = Std.int(y / GRID_SIZE);
		xr = (x - cx*GRID_SIZE) / GRID_SIZE;
		yr = (y - cy*GRID_SIZE) / GRID_SIZE;
		_needsUpdate();
	}

	public function setDestination(x:Float, y:Float, ?cb:Entity->Void = null):Void {
		hasDestination = true;
		destX = x;
		destY = y;
		destCallback = cb;
	}

	public inline function playAnimation(?name:String = ""):Bool {
		if (name == "") {
			current_animation_name = "";
			current_animation = null;
			current_step = 0;
			return false;
		} else if (animations == null || !animations.exists(name) || current_animation_name == name) {
			// current_animation_name = "";
			// current_animation = null;
			// current_step = 0;
			#if DEBUG
			GameConsole.log("Animation " + name + " not defined");
			#end
			return false;
		} else {
			current_animation_name = name;
			current_animation = animations.get(name);
			current_step = 0;
			return true;
		}
	}

	private function preUpdate():Void {}

	override private function onUpdate():Void {
		if (has_preupdate) preUpdate();

		// Enemies
		if (!isDeadEntity) {
			for (g in _foes) {
				var entities:Array<Entity> = pool.get(g);

				for (e in entities) {
					// Fast distance check
					if (e != this && !e.isDeadEntity && KMath.abs(cx-e.cx) <= 2 && KMath.abs(cy-e.cy) <= 2) {
						// Real distance check
						var dist:Float = Math.sqrt((e.x-x)*(e.x-x) + (e.y-y)*(e.y-y));

						if (dist <= radius+e.radius) {
							var ang:Float = Math.atan2(e.y-y, e.x-x);
							var force:Float = 0.2;
							var repelPower:Float = 0; //(radius+e.radius - dist) / (radius+e.radius);

							dx -= Math.cos(ang) * repelPower * force;
							dy -= Math.sin(ang) * repelPower * force;
							e.dx += Math.cos(ang) * repelPower * force;
							e.dy += Math.sin(ang) * repelPower * force;

							// Ignore damage when protected
							if (group == "player" && data == "player") {
								var player:Player = cast this;
								if (player.state == PROTECTED) continue;
							}

							// Ignore damage when protected
							if (e.group == "player" && e.data == "player") {
								var player:Player = cast e;
								if (player.state == PROTECTED) continue;
							}

							// Ignore offscreen player bullets
							if (group == "player-bullet" && y < 0) continue;
							if (e.group == "player-bullet" && e.y < 0) continue;

							dealDamage(e.attack, 0);
							e.dealDamage(attack, 0);
						}
					}
				}
			}
		}

		// Custom destination
		if (!isDeadEntity && hasDestination) {
			var threshold:Float = 2;
			var distX:Float = destX - x;
			var distY:Float = destY - y;
			if (Math.abs(distX) < threshold && Math.abs(distY) < threshold) {
				hasDestination = false;
				if (destCallback != null) destCallback(this);
			} else {
				var distTotal:Float = Math.abs(distX) + Math.abs(distY);
				var ratioX:Float = distX / distTotal;
				var ratioY:Float = distY / distTotal;

				var speed:Float = maxSpeed;
				if (group == "player" &&
					Main.nships == 2 &&
					Main.shipA.state == PROTECTED &&
					Main.shipB.state == PROTECTED
				) {
					speed *= 0.3;
				}

				dx = speed * ratioX;
				dy = speed * ratioY;
			}
		} else if (symetricFrom != null) {
			setCoordinates(Main.GAME_WIDTH - symetricFrom.x, symetricFrom.y);
		}

		// X component
		xr += dx;
		while (xr < 0) {
			cx--;
			xr++;
		}
		while (xr > 1) {
			cx++;
			xr--;
		}

		// Y component
		yr += dy;
		while (yr < 0) {
			cy--;
			yr++;
		}
		while (yr > 1) {
			cy++;
			yr--;
		}

		if (!needsUpdate) {
			var lastx:Int = x;
			var lasty:Int = y;

			x = Std.int((cx+xr)*GRID_SIZE);
			y = Std.int((cy+yr)*GRID_SIZE);

			if (x != lastx || y != lasty) _needsUpdate();
		} else {
			x = Std.int((cx+xr)*GRID_SIZE);
			y = Std.int((cy+yr)*GRID_SIZE);
		}

		// Get current tile
		if (current_animation != null && Main.kframe % 10 == 0) {
			var last_tile:String = tile_name;
			tile_name = current_animation[current_step];

			// Only reset tile data if something changed
			if (last_tile != tile_name) {
				tile_id = -1;
				_needsUpdate();
			}

			current_step++;
			if (current_step >= current_animation.length) current_step = 0;

		}

		lastcy = cy;
	}

	public function destroy():Void {
		if (isDeadEntity) return;

		isDeadEntity = true;
		pool.get(group).remove(this);
		_parent.removeEntity(this);
		_needsUpdate();
	}
}
